import uuid
from pathlib import Path
from typing import List, NamedTuple
from urllib.parse import unquote, urlparse

import mistune


class RelativeLink(NamedTuple):
    path: str
    fragment: str
    replacement: str
    original: str
    escaped_original: str


class ConfluenceTag(object):
    def __init__(self, name, text="", attrib=None, namespace="ac", cdata=False):
        self.name = name
        self.text = text
        self.namespace = namespace
        if attrib is None:
            attrib = {}
        self.attrib = attrib
        self.children = []
        self.cdata = cdata

    def render(self):
        namespaced_name = self.add_namespace(self.name, namespace=self.namespace)
        namespaced_attribs = {
            self.add_namespace(
                attribute_name, namespace=self.namespace
            ): attribute_value
            for attribute_name, attribute_value in self.attrib.items()
        }

        content = "<{}{}>{}{}</{}>".format(
            namespaced_name,
            " {}".format(
                " ".join(
                    [
                        '{}="{}"'.format(name, value)
                        for name, value in sorted(namespaced_attribs.items())
                    ]
                )
            )
            if namespaced_attribs
            else "",
            "".join([child.render() for child in self.children]),
            "<![CDATA[{}]]>".format(self.text) if self.cdata else self.text,
            namespaced_name,
        )
        return "{}\n".format(content)

    @staticmethod
    def add_namespace(tag, namespace):
        return "{}:{}".format(namespace, tag)

    def append(self, child):
        self.children.append(child)


class ConfluenceRenderer(mistune.HTMLRenderer):
    def __init__(
        self,
        strip_header=False,
        remove_text_newlines=False,
        enable_relative_links=False,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.strip_header = strip_header
        self.remove_text_newlines = remove_text_newlines
        self.attachments = list()
        self.title = None
        self.enable_relative_links = enable_relative_links
        self.relative_links: List[RelativeLink] = list()

    def reinit(self):
        self.attachments = list()
        self.relative_links = list()
        self.title = None

    def heading(self, text, level, raw=None):
        if self.title is None and level == 1:
            self.title = text
            # Don't duplicate page title as a header
            if self.strip_header:
                return ""

        return super(ConfluenceRenderer, self).heading(text, level, raw=raw)

    def structured_macro(self, name):
        return ConfluenceTag("structured-macro", attrib={"name": name})

    def parameter(self, name, value):
        parameter_tag = ConfluenceTag("parameter", attrib={"name": name})
        parameter_tag.text = value
        return parameter_tag

    def plain_text_body(self, text):
        body_tag = ConfluenceTag("plain-text-body", cdata=True)
        body_tag.text = text
        return body_tag

    def link(self, text, url, title=None):
        parsed_link = urlparse(url)
        if (
            self.enable_relative_links
            and (not parsed_link.scheme and not parsed_link.netloc)
            and parsed_link.path
        ):
            # relative link
            replacement_link = f"md2cf-internal-link-{uuid.uuid4()}"
            self.relative_links.append(
                RelativeLink(
                    # make sure to unquote the url as relative paths
                    # might have escape sequences
                    path=unquote(parsed_link.path),
                    replacement=replacement_link,
                    fragment=parsed_link.fragment,
                    original=url,
                    escaped_original=mistune.escape_url(url),
                )
            )
            url = replacement_link
        return super(ConfluenceRenderer, self).link(text, url, title)

    def text(self, text):
        if self.remove_text_newlines:
            text = text.replace("\n", " ")

        return text

    def block_code(self, code, lang=None):
        root_element = self.structured_macro("code")
        if lang is not None:
            lang_parameter = self.parameter(name="language", value=lang)
            root_element.append(lang_parameter)
        root_element.append(self.parameter(name="linenumbers", value="true"))
        root_element.append(self.plain_text_body(code))
        return root_element.render()

    def image(self, src, title, text):
        attributes = {"alt": text}
        if title:
            attributes["title"] = title

        root_element = ConfluenceTag(name="image", attrib=attributes)
        parsed_source = urlparse(src)
        if not parsed_source.netloc:
            # Local file, requires upload
            basename = Path(src).name
            url_tag = ConfluenceTag(
                "attachment", attrib={"filename": basename}, namespace="ri"
            )
            self.attachments.append(src)
        else:
            url_tag = ConfluenceTag("url", attrib={"value": src}, namespace="ri")
        root_element.append(url_tag)

        return root_element.render()



from typing import Dict

import rich.live
import rich.progress
import rich.table
import rich.text
import rich.tree

from schneiderl_md2cf.console_output import console


class Md2cfTUI(object):
    def __init__(self, pages_to_upload):
        tree = rich.tree.Tree("Pages to upload", hide_root=True)
        progress_table = rich.table.Table.grid()
        progress_table.row_styles = ["dim", ""]
        title_to_tree: Dict[str, rich.tree.Tree] = dict()
        self.title_to_progress: Dict[str, rich.progress.Progress] = dict()
        for page in pages_to_upload:
            if page.file_path is None and len(pages_to_upload) > 1:
                pretty_title = f":open_file_folder: {page.title}"
            else:
                pretty_title = f":page_facing_up: {page.title}"

            page_progress = rich.progress.Progress(
                rich.progress.BarColumn(),
                rich.progress.SpinnerColumn(finished_text=""),
                rich.progress.TextColumn(""),
                console=console,
            )
            page_progress.add_task(
                description="", total=1 + len(page.attachments), start=False
            )
            self.title_to_progress[page.title] = page_progress

            if page.parent_title:
                try:
                    page_node = title_to_tree[page.parent_title].add(
                        pretty_title, style="bright"
                    )
                except KeyError:
                    continue
            else:
                page_node = tree.add(pretty_title, style="bright")

            progress_table.add_row(page_progress)
            title_to_tree[page.title] = page_node

            for attachment in page.attachments:
                page_node.add(f":paperclip: {attachment}", style="dim")
                attachment_progress = rich.progress.Progress(
                    rich.progress.BarColumn(),
                    rich.progress.SpinnerColumn(finished_text="done"),
                    rich.progress.TextColumn(""),
                    console=console,
                )
                attachment_progress.add_task(description="", total=1, start=False)
                progress_table.add_row(attachment_progress)
                self.title_to_progress[
                    f"{page.title} {attachment}"
                ] = attachment_progress
        self.overall_progress = rich.progress.Progress(console=console)
        self.overall_progress.add_task(
            "Total progress",
            start=True,
            total=len(pages_to_upload)
            + sum([len(page.attachments) for page in pages_to_upload]),
        )
        table = rich.table.Table().grid(padding=1, pad_edge=True)
        tree_and_progress = rich.table.Table.grid(
            "", rich.table.Column(""), expand=True
        )
        tree_and_progress.add_row(tree, progress_table)
        table.add_row(tree_and_progress)
        table.add_row(self.overall_progress)
        self.live = rich.live.Live(table, console=console, refresh_per_second=10)
        # return self.overall_progress, overall_task, table, self.title_to_progress

    def __enter__(self):
        self.live.__enter__()
        return self

    def __exit__(
        self,
        *args,
        **kwargs,
    ):
        self.live.__exit__(*args, **kwargs)

    def set_item_progress_label(self, item_name, label):
        self.title_to_progress[item_name].columns[2].text_format = label

    def set_item_finished_text(self, item_name, finished_text):
        self.title_to_progress[item_name].columns[1].finished_text = finished_text

    def set_item_finished_text_from_result(self, item_name, upsert_result):
        self.set_item_finished_text(
            item_name, Md2cfTUI.format_upsert_result(upsert_result)
        )

    def tick_item_progress(self, item_name):
        task_for_this_item = self.title_to_progress[item_name].task_ids[0]
        self.title_to_progress[item_name].update(
            task_id=task_for_this_item,
            advance=1,
        )

    def tick_global_progress(self):
        self.overall_progress.update(
            task_id=self.overall_progress.task_ids[0], advance=1
        )

    def start_item_task(self, item_name):
        task_for_this_item = self.title_to_progress[item_name].task_ids[0]
        self.title_to_progress[item_name].start_task(task_id=task_for_this_item)

    def reset_item_task(self, item_name, total):
        task_for_this_item = self.title_to_progress[item_name].task_ids[0]
        self.title_to_progress[item_name].reset(task_id=task_for_this_item, total=total)

    @staticmethod
    def format_upsert_result(upsert_item_result):
        if upsert_item_result.action == upsert_item_result.action.CREATED:
            finished_text = rich.text.Text.from_markup(
                "[green]:heavy_check_mark-emoji: Created"
            )
        elif upsert_item_result.action == upsert_item_result.action.UPDATED:
            finished_text = rich.text.Text.from_markup(
                "[green]:heavy_check_mark-emoji: Updated"
            )
        elif upsert_item_result.action == upsert_item_result.action.SKIPPED:
            finished_text = rich.text.Text.from_markup("[yellow]No change")

        return finished_text



from urllib.parse import urljoin

import requests
import requests.adapters
import requests.packages
import urllib3


def bunchify(obj):
    if isinstance(obj, (list, tuple)):
        return [bunchify(item) for item in obj]
    if isinstance(obj, dict):
        return Bunch(obj)
    return obj


class Bunch(dict):
    def __init__(self, kwargs=None):
        if kwargs is None:
            kwargs = {}
        for key, value in kwargs.items():
            kwargs[key] = bunchify(value)
        super(Bunch, self).__init__(kwargs)
        self.__dict__ = self


class MinimalConfluence:
    def __init__(
        self, host, username=None, password=None, token=None, verify=True, max_retries=4
    ):
        if token is None:
            if username is None and password is None:
                raise ValueError(
                    "Either a personal access token, "
                    "or username and password are required"
                )

        if not host.endswith("/"):
            self.host = host + "/"
        else:
            self.host = host
        self.api = requests.Session()
        self.api.verify = verify

        if token is not None:
            self.api.headers.update({"Authorization": f"Bearer {token}"})
        elif username is not None and password is not None:
            self.api.auth = (username, password)

        adapter = requests.adapters.HTTPAdapter(
            max_retries=urllib3.Retry(
                total=max_retries,
                backoff_factor=1,
                respect_retry_after_header=True,
                allowed_methods=None,
            )
        )
        self.api.mount("http://", adapter)
        self.api.mount("https://", adapter)

        if not verify:
            requests.packages.urllib3.disable_warnings(
                requests.packages.urllib3.exceptions.InsecureRequestWarning
            )

    def _request(self, method, path, **kwargs):
        r = self.api.request(method, urljoin(self.host, path), **kwargs)
        r.raise_for_status()
        return bunchify(r.json())

    def _get(self, path, **kwargs):
        return self._request("GET", path, **kwargs)

    def _post(self, path, **kwargs):
        return self._request("POST", path, **kwargs)

    def _put(self, path, **kwargs):
        return self._request("PUT", path, **kwargs)

    def get_page(
        self,
        title=None,
        space_key=None,
        page_id=None,
        content_type="page",
        additional_expansions=None,
    ):
        """
        Create a new page in a space

        Args:
            title (str): the title for the page
            space_key (str): the Confluence space for the page
            content_type (str): Content type. Default value: page.
              Valid values: page, blogpost.
            page_id (str or int): the ID of the page
            additional_expansions (list of str): Additional expansions that should be
              made when calling the api

        Returns:
            The response from the API

        """
        params = None
        if additional_expansions is not None:
            params = {"expand": ",".join(additional_expansions)}

        if page_id is not None:
            return self._get(f"content/{page_id}", params=params)
        elif title is not None:
            params = {"title": title, "type": content_type}
            if space_key is not None:
                params["spaceKey"] = space_key
            response = self._get("content", params=params)
            try:
                # A search by title/space doesn't return full page objects,
                # and since we don't support expansion in this implementation
                # just yet, we just retrieve the "full" page data using the page
                # ID for the first search result
                return self.get_page(
                    page_id=response.results[0].id,
                    additional_expansions=additional_expansions,
                )
            except IndexError:
                return None
        else:
            raise ValueError("At least one of title or page_id must not be None")

    def create_page(
        self,
        space,
        title,
        body,
        content_type="page",
        parent_id=None,
        update_message=None,
        labels=None,
    ):
        """
        Create a new page in a space

        Args:
            space (str): the Confluence space for the page
            title (str): the title for the page
            body (str): the body of the page, in Confluence Storage Format
            content_type (str): Content type. Default value: page.
              Valid values: page, blogpost.
            parent_id (str or int): the ID of the parent page
            update_message (str): optional. A message that will appear in Confluence's
              history
            labels (list(str)): optional. The set of labels the final page should have.
              None leaves existing labels unchanged

        Returns:
            The response from the API

        """
        page_structure = {
            "title": title,
            "type": content_type,
            "space": {"key": space},
            "body": {"storage": {"value": body, "representation": "storage"}},
        }

        if parent_id is not None:
            if type(parent_id) is str:
                parent_id = int(parent_id)
            page_structure["ancestors"] = [{"id": parent_id}]

        if update_message is not None:
            page_structure["version"] = {"message": update_message}

        if labels is not None:
            page_structure["metadata"] = {
                "labels": [{"name": label, "prefix": "global"} for label in labels]
            }

        return self._post("content", json=page_structure)

    def update_page(
        self,
        page,
        body,
        parent_id=None,
        content_type="page",
        update_message=None,
        labels=None,
        minor_edit=False,
    ):
        update_structure = {
            "version": {
                "number": page.version.number + 1,
                "minorEdit": minor_edit,
            },
            "title": page.title,
            "type": content_type,
            "body": {"storage": {"value": body, "representation": "storage"}},
        }

        if parent_id is not None:
            if type(parent_id) is str:
                parent_id = int(parent_id)
            update_structure["ancestors"] = [{"id": parent_id}]

        if update_message is not None:
            update_structure["version"]["message"] = update_message

        if labels is not None:
            update_structure["metadata"] = {
                "labels": [{"name": label, "prefix": "global"} for label in labels]
            }

        return self._put(f"content/{page.id}", json=update_structure)

    def get_attachment(self, confluence_page, name):
        existing_attachments = self._get(
            f"content/{confluence_page.id}/child/attachment",
            headers={"X-Atlassian-Token": "nocheck", "Accept": "application/json"},
            params={"filename": name, "expand": "version"},
        )

        if existing_attachments.size:
            return existing_attachments.results[0]

    def update_attachment(self, confluence_page, fp, existing_attachment, message=""):
        return self._post(
            f"content/{confluence_page.id}/child/attachment/{existing_attachment.id}/"
            f"data",
            headers={"X-Atlassian-Token": "nocheck"},
            files={"file": fp, "comment": message if message else None},
        )

    def create_attachment(self, confluence_page, fp, message=""):
        return self._post(
            f"content/{confluence_page.id}/child/attachment",
            json={"comment": message} if message else None,
            headers={"X-Atlassian-Token": "nocheck"},
            params={"allowDuplicated": "true"},
            files={"file": fp, "comment": message if message else None},
        )

    def add_labels(self, page, labels):
        # return self.api.content(page.id).post(
        return self._post(
            f"content/{page.id}/label",
            data=[{"name": label, "type": "global"} for label in labels],
        )

    def get_url(self, page):
        return f"{page._links.base}{page._links.webui}"

    def get_parent_id(self, page):
        return page.ancestors[-1].id

    def get_space(self, space, additional_expansions=None):
        params = None
        if additional_expansions is not None:
            params = {"expand": ",".join(additional_expansions)}
        return self._get(f"space/{space}", params=params)



import hashlib
import os
from pathlib import Path
from typing import Any, Dict, List, Optional

import chardet
import mistune
import yaml
from yaml.parser import ParserError

from schneiderl_md2cf.confluence_renderer import ConfluenceRenderer, RelativeLink
from schneiderl_md2cf.ignored_files import GitRepository


class Page(object):
    def __init__(
        self,
        title: Optional[str],
        body: str,
        content_type: Optional[str] = "page",
        attachments: Optional[List[Path]] = None,
        file_path: Optional[Path] = None,
        page_id: str = None,
        parent_id: str = None,
        parent_title: str = None,
        space: str = None,
        labels: Optional[List[str]] = None,
        relative_links: Optional[List[RelativeLink]] = None,
    ):
        self.title = title
        self.original_title = None
        self.body = body
        self.content_type = content_type
        self.file_path = file_path
        self.attachments = attachments
        if self.attachments is None:
            self.attachments: List[Path] = list()
        self.relative_links = relative_links
        if self.relative_links is None:
            self.relative_links: List[RelativeLink] = list()
        self.page_id = page_id
        self.parent_id = parent_id
        self.parent_title = parent_title
        self.space = space
        self.labels = labels

    def get_content_hash(self):
        return hashlib.sha1(self.body.encode()).hexdigest()

    def __repr__(self):
        return "Page({})".format(
            ", ".join(
                [
                    "{}={}".format(name, repr(value))
                    for name, value in [
                        ["title", self.title],
                        ["file_path", self.file_path],
                        ["page_id", self.page_id],
                        ["parent_id", self.parent_id],
                        ["parent_title", self.parent_title],
                        ["space", self.space],
                        [
                            "body",
                            f"{self.body[:40]} [...]"
                            if len(self.body) > 40
                            else self.body,
                        ],
                    ]
                ]
            )
        )


def find_non_empty_parent_path(
    current_dir: Path, folder_data: Dict[Path, Dict[str, Any]], default: Path
) -> Path:
    for parent in current_dir.parents:
        if parent in folder_data and folder_data[parent]["n_files"]:
            return parent
    return default.absolute()


def get_pages_from_directory(
    file_path: Path,
    collapse_single_pages: bool = False,
    skip_empty: bool = False,
    collapse_empty: bool = False,
    beautify_folders: bool = False,
    use_pages_file: bool = False,
    strip_header: bool = False,
    remove_text_newlines: bool = False,
    use_gitignore: bool = True,
    enable_relative_links: bool = False,
) -> List[Page]:
    """
    Collect a list of markdown files recursively under the file_path directory.

    :param file_path: The starting path from which to search
    :param collapse_single_pages:
    :param skip_empty:
    :param collapse_empty:
    :param beautify_folders:
    :param use_pages_file:
    :param strip_header:
    :param remove_text_newlines:
    :param use_gitignore: Use .gitignore files to skip unwanted markdown in directory
      search
    :param enable_relative_links: extract all relative links and replace them with
      placeholders
    :return: A list of paths to the markdown files to upload.
    """
    processed_pages = list()
    base_path = file_path.resolve()
    folder_data = dict()
    git_repo = GitRepository(file_path, use_gitignore=use_gitignore)

    for current_path, directories, file_names in os.walk(file_path):
        current_path = Path(current_path).resolve()

        if git_repo.is_ignored(current_path):
            continue

        markdown_files = [
            Path(current_path, file_name)
            for file_name in file_names
            if file_name.endswith(".md")
        ]
        # Filter out ignored files
        markdown_files = [
            path for path in markdown_files if not git_repo.is_ignored(path)
        ]

        folder_data[current_path] = {"n_files": len(markdown_files)}

        # we'll capture title and path of the parent folder for this folder:
        folder_parent_title = None
        folder_parent_path = None

        # title for this folder's page (as parent of its children):
        parent_page_title = None
        # title for the folder (same as above except when collapsing):
        folder_title = None

        if current_path != base_path:
            # TODO: add support for .pages file to read folder title
            if skip_empty or collapse_empty:
                folder_parent_path = find_non_empty_parent_path(
                    current_path, folder_data, default=file_path
                )
            else:
                folder_parent_path = current_path.parent

            folder_parent_title = folder_data[folder_parent_path]["title"]
            parent_page_title = current_path.name
            if len(markdown_files) == 1 and collapse_single_pages:
                parent_page_title = folder_parent_title
                folder_title = None
            else:
                if collapse_empty:
                    parent_page_title = str(
                        current_path.relative_to(folder_parent_path)
                    )
                if beautify_folders:
                    parent_page_title = (
                        current_path.name.replace("-", " ")
                        .replace("_", " ")
                        .capitalize()
                    )
                folder_title = parent_page_title
        if use_pages_file and ".pages" in file_names:
            with open(current_path.joinpath(".pages")) as pages_fp:
                pages_file_contents = yaml.safe_load(pages_fp)
            if "title" in pages_file_contents:
                parent_page_title = pages_file_contents["title"]
                folder_title = parent_page_title

        folder_data[current_path]["title"] = folder_title

        if folder_title is not None and (
            markdown_files or (directories and not skip_empty and not collapse_empty)
        ):
            processed_pages.append(
                Page(
                    title=folder_title,
                    parent_title=folder_parent_title,
                    body="",
                )
            )

        for markdown_file in markdown_files:
            processed_page = get_page_data_from_file_path(
                markdown_file,
                strip_header=strip_header,
                remove_text_newlines=remove_text_newlines,
                enable_relative_links=enable_relative_links,
            )
            processed_page.parent_title = parent_page_title
            processed_pages.append(processed_page)

            # This replaces the title for the current folder with the title for the
            # document we just parsed, so things below this folder will be correctly
            # parented to the collapsed document.
            if len(markdown_files) == 1 and collapse_single_pages:
                folder_data[current_path]["title"] = processed_page.title

    return processed_pages


def get_page_data_from_file_path(
    file_path: Path,
    strip_header: bool = False,
    remove_text_newlines: bool = False,
    enable_relative_links: bool = False,
) -> Page:
    if not isinstance(file_path, Path):
        file_path = Path(file_path)

    try:
        with open(file_path) as file_handle:
            markdown_lines = file_handle.readlines()
    except UnicodeDecodeError:
        with open(file_path, "rb") as file_handle:
            detected_encoding = chardet.detect(file_handle.read())
        with open(file_path, encoding=detected_encoding["encoding"]) as file_handle:
            markdown_lines = file_handle.readlines()

    page = get_page_data_from_lines(
        markdown_lines,
        strip_header=strip_header,
        remove_text_newlines=remove_text_newlines,
        enable_relative_links=enable_relative_links,
    )

    if not page.title:
        page.title = file_path.stem

    page.file_path = file_path

    return page


def get_page_data_from_lines(
    markdown_lines: List[str],
    strip_header: bool = False,
    remove_text_newlines: bool = False,
    enable_relative_links: bool = False,
) -> Page:
    frontmatter = get_document_frontmatter(markdown_lines)
    if "frontmatter_end_line" in frontmatter:
        markdown_lines = markdown_lines[frontmatter["frontmatter_end_line"] :]

    page = parse_page(
        markdown_lines,
        strip_header=strip_header,
        remove_text_newlines=remove_text_newlines,
        enable_relative_links=enable_relative_links,
    )

    if "title" in frontmatter:
        page.title = frontmatter["title"]

    if "labels" in frontmatter:
        if isinstance(frontmatter["labels"], list):
            page.labels = [str(label) for label in frontmatter["labels"]]
        else:
            raise TypeError(
                "the labels section in the frontmatter " "must be a list of strings"
            )
    return page


def parse_page(
    markdown_lines: List[str],
    strip_header: bool = False,
    remove_text_newlines: bool = False,
    enable_relative_links: bool = False,
    escape: bool = False,
    allow_harmful_protocols: bool = False,
) -> Page:
    renderer = ConfluenceRenderer(
        strip_header=strip_header,
        remove_text_newlines=remove_text_newlines,
        enable_relative_links=enable_relative_links,
        escape=escape,
        allow_harmful_protocols=allow_harmful_protocols,
    )
    confluence_mistune = mistune.Markdown(renderer=renderer)
    confluence_content = confluence_mistune("".join(markdown_lines))

    page = Page(
        title=renderer.title,
        body=confluence_content,
        attachments=renderer.attachments,
        relative_links=renderer.relative_links,
    )

    return page


def get_document_frontmatter(markdown_lines: List[str]) -> Dict[str, Any]:
    frontmatter_yaml = ""
    frontmatter_end_line = 0
    if markdown_lines and markdown_lines[0] == "---\n":
        for index, line in enumerate(markdown_lines[1:]):
            if line == "---\n":
                frontmatter_end_line = index + 2
                break
            else:
                frontmatter_yaml += line
    frontmatter = None
    if frontmatter_yaml and frontmatter_end_line:
        try:
            frontmatter = yaml.safe_load(frontmatter_yaml)
        except ParserError:
            pass
    if isinstance(frontmatter, dict):
        frontmatter["frontmatter_end_line"] = frontmatter_end_line
    else:
        frontmatter = {}

    return frontmatter



import argparse
import copy
import getpass
import os
import sys
from collections import Counter
from pathlib import Path
from typing import List

import rich.table
import rich.text
import rich.tree
from requests import HTTPError
from rich import box
from rich_argparse import RichHelpFormatter

import schneiderl_md2cf.document
from schneiderl_md2cf import api
from schneiderl_md2cf.console_output import (
    console,
    error_console,
    json_output_console,
    minimal_output_console,
)
from schneiderl_md2cf.document import Page
from schneiderl_md2cf.tui import schneiderl_md2cfTUI
from schneiderl_md2cf.upsert import upsert_attachment, upsert_page


def get_parser():
    parser = argparse.ArgumentParser(formatter_class=RichHelpFormatter)
    login_group = parser.add_argument_group("login arguments")
    login_group.add_argument(
        "-o",
        "--host",
        help="full URL of the Confluence instance. "
        "Can also be specified as CONFLUENCE_HOST environment variable.",
        default=os.getenv("CONFLUENCE_HOST"),
    )
    login_group.add_argument(
        "-u",
        "--username",
        help="username for logging into Confluence. "
        "Can also be specified as CONFLUENCE_USERNAME environment variable.",
        default=os.getenv("CONFLUENCE_USERNAME"),
    )
    login_group.add_argument(
        "-p",
        "--password",
        help="password for logging into Confluence. "
        "Can also be specified as CONFLUENCE_PASSWORD environment variable. "
        "If not specified, it will be asked for interactively.",
        default=os.getenv("CONFLUENCE_PASSWORD"),
    )
    login_group.add_argument(
        "--token",
        help="personal access token for logging into Confluence. "
        "Can also be specified as CONFLUENCE_TOKEN environment variable.",
        default=os.getenv("CONFLUENCE_TOKEN"),
    )
    login_group.add_argument(
        "--insecure",
        action="store_true",
        help="do not verify SSL certificates",
    )

    required_group = parser.add_argument_group("required arguments")
    required_group.add_argument(
        "-s",
        "--space",
        help="key for the Confluence space the page will be published to. "
        "Can also be specified as CONFLUENCE_SPACE environment variable.",
        default=os.getenv("CONFLUENCE_SPACE"),
    )

    output_group = parser.add_argument_group("schneiderl_md2cf output arguments")
    output_group.add_argument(
        "--output",
        choices=["default", "minimal", "json"],
        default="default",
    )

    page_group = parser.add_argument_group("page information arguments")
    parent_group = page_group.add_mutually_exclusive_group()
    parent_group.add_argument(
        "-a",
        "--parent-title",
        help="title of the parent page under which the new page will be uploaded",
    )
    parent_group.add_argument(
        "-A",
        "--parent-id",
        help="ID of the parent page under which the new page will be uploaded",
    )
    parent_group.add_argument(
        "--top-level",
        action="store_true",
        help="upload the page tree starting from the top level (no top level parent)",
    )

    page_group.add_argument(
        "-t",
        "--title",
        help="a title for the page. Determined from the document if missing",
    )

    page_group.add_argument(
        "-c",
        "--content-type",
        help="Content type. Default value: page",
        choices=["page", "blogpost"],
        default="page",
    )

    page_group.add_argument("-m", "--message", help="update message for the change")
    page_group.add_argument(
        "--minor-edit", action="store_true", help="do not notify watchers of change"
    )
    page_group.add_argument("-i", "--page-id", help="ID of the page to be updated")
    page_group.add_argument(
        "--prefix",
        help="a string to prefix to every page title to ensure uniqueness",
        type=str,
    )
    page_group.add_argument(
        "--strip-top-header",
        action="store_true",
        help="remove the top level header from the page",
    )
    page_group.add_argument(
        "--remove-text-newlines",
        action="store_true",
        help="remove single newlines in paragraphs",
    )
    page_group.add_argument(
        "--replace-all-labels",
        action="store_true",
        help="replace all labels instead of only adding new ones",
    )

    preface_group = page_group.add_mutually_exclusive_group()
    preface_group.add_argument(
        "--preface-markdown",
        nargs="?",
        type=str,
        default=None,
        const="**Contents are auto-generated, do not edit.**",
        help="markdown content to prepend to each page. "
        'Defaults to "**Contents are auto-generated, do not edit.**" '
        "if no markdown is specified",
    )
    preface_group.add_argument(
        "--preface-file",
        type=Path,
        help="path to a markdown file to be prepended to every page",
    )

    postface_group = page_group.add_mutually_exclusive_group()
    postface_group.add_argument(
        "--postface-markdown",
        nargs="?",
        type=str,
        default=None,
        const="**Contents are auto-generated, do not edit.**",
        help="markdown content to append to each page. "
        'Defaults to "**Contents are auto-generated, do not edit.**" '
        "if no markdown is specified",
    )
    postface_group.add_argument(
        "--postface-file",
        type=Path,
        help="path to a markdown file to be appended to every page",
    )

    dir_group = parser.add_argument_group("directory arguments")
    dir_group.add_argument(
        "--collapse-single-pages",
        action="store_true",
        help="if a folder contains a single document, collapse it "
        "so the folder doesn't appear",
    )
    dir_group.add_argument(
        "--no-gitignore",
        action="store_false",
        dest="use_gitignore",
        default=True,
        help="do not use .gitignore files to filter directory search",
    )
    dir_title_group = dir_group.add_mutually_exclusive_group()
    dir_title_group.add_argument(
        "--beautify-folders",
        action="store_true",
        help="replace hyphens and underscore in folder names with spaces, "
        "and capitalize the first letter",
    )
    dir_title_group.add_argument(
        "--use-pages-file",
        action="store_true",
        help='use the "title" entry in YAML files called .pages in each '
        "directory to change the folder name",
    )

    empty_group = dir_group.add_mutually_exclusive_group()
    empty_group.add_argument(
        "--collapse-empty",
        action="store_true",
        help="collapse multiple empty folders into one",
    )
    empty_group.add_argument(
        "--skip-empty",
        action="store_true",
        help="if a folder doesn't contain documents, skip it",
    )

    relative_links_group = parser.add_argument_group("relative links arguments")
    relative_links_group.add_argument(
        "--enable-relative-links",
        action="store_true",
        help="enable parsing of relative links to other markdown files. "
        "Requires two passes for pages with relative links, and will cause them "
        "to always be updated regardless of the --only-changed flag",
    )
    relative_links_group.add_argument(
        "--ignore-relative-link-errors",
        action="store_true",
        help="when relative links are enabled and a link doesn't point to an "
        "existing and uploaded file, leave the link as-is instead of exiting.",
    )

    parser.add_argument(
        "--dry-run",
        action="store_true",
        help="print information on all the pages instead of uploading to Confluence",
    )
    parser.add_argument(
        "--debug",
        action="store_true",
        help="print full stack traces for exceptions",
    )
    parser.add_argument(
        "--only-changed",
        action="store_true",
        help="only upload pages and attachments that have changed. "
        "This adds a hash of the page or attachment contents to the update message",
    )
    parser.add_argument(
        "file_list",
        type=Path,
        help="markdown files or directories to upload to Confluence. Empty for stdin",
        nargs="*",
    )
    parser.add_argument(
        "--max-retries",
        action="store",
        dest="max_retries",
        type=int,
        default=4,
        help="number of retry attempts if any API call fails",
    )

    return parser


def print_missing_parameter(parameter_name: str):
    error_console.log(
        ":x: Missing required parameter: {}\n"
        "Use {} --help to get help.".format(parameter_name, sys.argv[0])
    )


def print_page_details(page: Page):
    page_copy = copy.copy(page)
    page_copy.body = page_copy.body[:40] + ("..." if page_copy.body else "")
    console.print(page_copy)


def main():
    args = get_parser().parse_args()

    if args.host is None:
        print_missing_parameter("host")
        sys.exit(1)

    if args.username is None and args.token is None:
        print_missing_parameter("username or bearer token")
        sys.exit(1)

    if args.space is None:
        print_missing_parameter("space")
        sys.exit(1)

    if args.password is None and args.token is None:
        args.password = getpass.getpass()

    if args.output == "minimal":
        console.quiet = True
        minimal_output_console.quiet = False
    elif args.output == "json":
        console.quiet = True
        json_output_console.quiet = False

    confluence = api.MinimalConfluence(
        host=args.host,
        username=args.username,
        password=args.password,
        token=args.token,
        verify=not args.insecure,
        max_retries=args.max_retries,
    )

    if (args.title or args.page_id) and (
        len(args.file_list) > 1 or any(map(os.path.isdir, args.file_list))
    ):
        error_console.log(
            ":x: Title and page ID cannot be specified on the command line "
            "if uploading more than one file or whole directories\n"
        )
        sys.exit(1)

    pages_to_upload = collect_pages_to_upload(args)

    page_title_counts = Counter([page.title for page in pages_to_upload])
    colliding_titles = [
        title for title, count in page_title_counts.most_common() if count > 1
    ]
    if colliding_titles:
        error_console.log(
            ":x: Some documents have the [bold]same title[/], but all Confluence pages "
            "in the same space must have different titles.\n\n"
            "These are the documents (and path, if available) with identical titles:",
            markup=True,
        )
        colliding_titles_table = rich.table.Table(box=box.SIMPLE)
        colliding_titles_table.add_column("Title")
        colliding_titles_table.add_column("File")
        for title in colliding_titles:
            for filename in [
                page.file_path for page in pages_to_upload if page.title == title
            ]:
                # error_console.log(f"{title}\t{filename}\n", markup=True)
                colliding_titles_table.add_row(title, str(filename))
        error_console.log(colliding_titles_table)
        sys.exit(1)

    for page in pages_to_upload:
        for attachment in page.attachments:
            if page.file_path is not None:
                attachment_path = page.file_path.parent.joinpath(attachment)
            else:
                attachment_path = attachment

            if not attachment_path.is_file():
                error_console.log(
                    f"[bold red]:x: ERROR:[default] attachment {attachment_path} "
                    f"for page {page.title} does not exist"
                )
                sys.exit(1)

    preface_markup = ""
    if args.preface_markdown:
        preface_markup = schneiderl_md2cf.document.parse_page([args.preface_markdown]).body
    elif args.preface_file:
        # We don't use strip_header or remove_text_newlines here
        # since this is just a preface doc
        preface_markup = schneiderl_md2cf.document.get_page_data_from_file_path(
            args.preface_file
        ).body

    postface_markup = ""
    if args.postface_markdown:
        postface_markup = schneiderl_md2cf.document.parse_page([args.postface_markdown]).body
    elif args.postface_file:
        # We don't use strip_header or remove_text_newlines here
        # since this is just a postface doc
        postface_markup = schneiderl_md2cf.document.get_page_data_from_file_path(
            args.postface_file
        ).body

    map_document_path_to_confluence_page = dict()
    if args.enable_relative_links:
        map_document_path_to_confluence_page = build_document_path_to_page_map(
            pages_to_upload
        )
        if not args.ignore_relative_link_errors:
            validate_relative_links(
                pages_to_upload, map_document_path_to_confluence_page
            )

    something_went_wrong = False
    error = None
    tui = schneiderl_md2cfTUI(pages_to_upload)

    with tui:
        space_info = confluence.get_space(
            args.space, additional_expansions=["homepage"]
        )

        for page in pages_to_upload:
            pre_process_page(page, args, postface_markup, preface_markup, space_info)
            tui.start_item_task(page.original_title)
            upsert_page_result = None
            try:
                tui.set_item_progress_label(page.original_title, "Upserting")
                final_page = None
                if not args.dry_run:
                    upsert_page_result = upsert_page(
                        confluence=confluence,
                        message=args.message,
                        page=page,
                        only_changed=args.only_changed,
                        replace_all_labels=args.replace_all_labels,
                        minor_edit=args.minor_edit,
                    )
                    final_page = upsert_page_result.response
                    minimal_output_console.log(confluence.get_url(final_page))
                    json_output_console.print_json(data=final_page, indent=None)
                if page.attachments:
                    tui.set_item_progress_label(
                        page.original_title, "Processing attachments"
                    )
                    for attachment in page.attachments:
                        attachment_identifier = f"{page.original_title} {attachment}"
                        tui.start_item_task(attachment_identifier)
                        if not args.dry_run:
                            upsert_attachment_result = upsert_attachment(
                                confluence=confluence,
                                attachment=attachment,
                                existing_page=final_page,
                                message=args.message,
                                only_changed=args.only_changed,
                                page=page,
                            )
                            tui.set_item_finished_text_from_result(
                                attachment_identifier, upsert_attachment_result
                            )
                        else:
                            tui.set_item_finished_text(
                                attachment_identifier, "[yellow]Skipped (dry run)"
                            )
                        tui.set_item_progress_label(attachment_identifier, "")
                        tui.tick_item_progress(attachment_identifier)
                        tui.tick_item_progress(page.original_title)
                        tui.tick_global_progress()
                if page.file_path is not None and args.enable_relative_links:
                    # Skip pages without a file_path
                    # (e.g. section pages representing directories)
                    map_document_path_to_confluence_page[
                        page.file_path.resolve()
                    ] = final_page
            except HTTPError as e:
                if args.debug:
                    console.print_exception(show_locals=True)
                error = "{} - {}".format(str(e), e.response.content)
                something_went_wrong = True
            except Exception as e:
                if args.debug:
                    console.print_exception(show_locals=True)
                error = "[red]ERROR:[default] {}".format(str(e))
                something_went_wrong = True

            tui.set_item_progress_label(page.original_title, "")
            if not args.dry_run:
                if not something_went_wrong:
                    tui.set_item_finished_text_from_result(
                        page.original_title, upsert_page_result
                    )
                else:
                    tui.set_item_progress_label(
                        page.original_title, "[red]:x: Error while uploading"
                    )
            else:
                tui.set_item_finished_text(
                    page.original_title,
                    rich.text.Text.from_markup("[yellow]Skipped (dry run)"),
                )

            tui.tick_item_progress(page.original_title)
            tui.tick_global_progress()

            if something_went_wrong:
                break

        if not something_went_wrong and args.enable_relative_links:
            try:
                update_pages_with_relative_links(
                    args,
                    confluence,
                    pages_to_upload,
                    map_document_path_to_confluence_page,
                    tui,
                )
            except HTTPError as e:
                if args.debug:
                    console.print_exception(show_locals=True)
                error = "{} - {}".format(str(e), e.response.content)
                something_went_wrong = True
            except Exception as e:
                if args.debug:
                    console.print_exception(show_locals=True)
                error = "[red]ERROR:[default] {}".format(str(e))
                something_went_wrong = True

    if something_went_wrong:
        error_console.log(error)
        sys.exit(1)


def pre_process_page(page, args, postface_markup, preface_markup, space_info):
    page.original_title = page.title
    page.space = args.space
    page.page_id = args.page_id
    page.content_type = args.content_type

    if page.parent_title is None:  # This only happens for top level pages
        # If the argument is not supplied this leaves
        # the parent_title as None, which is fine
        page.parent_title = args.parent_title
    else:
        if args.prefix:
            page.parent_title = f"{args.prefix} - {page.parent_title}"

    if page.parent_title is None:
        page.parent_id = (
            page.parent_id or args.parent_id
        )  # This can still end up being None.
        # It's fine -- it means it's a top level page.

    # If we want to *move* a page back to the top space, we need to make it
    # a child of the space's home page
    if args.top_level and page.parent_title is None and page.parent_id is None:
        page.parent_id = space_info.homepage.id

    if args.prefix:
        page.title = f"{args.prefix} - {page.title}"

    if preface_markup:
        page.body = preface_markup + page.body

    if postface_markup:
        page.body = page.body + postface_markup


def validate_relative_links(pages_to_upload, path_to_page):
    invalid_links = False
    for page in pages_to_upload:
        for link_data in page.relative_links:
            link_absolute_path = (
                page.file_path.parent / Path(link_data.path)
            ).resolve()
            if link_absolute_path not in path_to_page:
                error_console.log(
                    f"Page {page.file_path} has a relative link to {link_data.path}"
                    ", which is not in the list of pages to be uploaded.\n"
                )
                invalid_links = True
    if invalid_links:
        error_console.log(
            "\nSome of the pages to be uploaded have invalid relative links.\n"
        )
        sys.exit(1)


def build_document_path_to_page_map(pages_to_upload):
    path_to_page = dict()
    for page in pages_to_upload:
        try:
            # Will be filled in later with the page returned by upsert
            path_to_page[page.file_path.resolve()] = None
        except AttributeError:
            # A page might not have a file_path
            # (for example if it's representing a directory)
            continue
    return path_to_page


def update_pages_with_relative_links(
    args, confluence, pages_to_upload, path_to_page, tui
):
    something_went_wrong = False
    error = ""
    for page in pages_to_upload:
        if page.file_path is None:
            # Skip pages without a file_path
            # (e.g. section pages representing directories)
            continue

        page_modified = False
        for link_data in page.relative_links:
            try:
                link_absolute_path = (
                    page.file_path.parent / Path(link_data.path)
                ).resolve()
                page_on_confluence = path_to_page[link_absolute_path]
            except KeyError:
                if args.ignore_relative_link_errors:
                    page.body = page.body.replace(
                        link_data.replacement,
                        link_data.escaped_original
                        + (("#" + link_data.fragment) if link_data.fragment else ""),
                    )
                    continue
                else:
                    error_console.log(
                        f"Page {page.file_path} has a relative link to {link_data.path}"
                        ", which was not uploaded correctly.\n"
                    )
                    break

            # in a dry run we don't actually have page URLs since we never upload
            # anything
            if not args.dry_run:
                page.body = page.body.replace(
                    link_data.replacement,
                    confluence.get_url(page_on_confluence)
                    + (("#" + link_data.fragment) if link_data.fragment else ""),
                )
            page_modified = True

        if page_modified:
            tui.reset_item_task(page.original_title, total=1)
            tui.set_item_progress_label(page.original_title, "Updating relative links")
            tui.start_item_task(page.original_title)
            if not args.dry_run:
                try:
                    upsert_page(
                        confluence=confluence,
                        message=args.message,
                        page=page,
                        only_changed=args.only_changed,
                        replace_all_labels=args.replace_all_labels,
                        minor_edit=True,
                    )
                except Exception as e:
                    error = e
                    something_went_wrong = True

                if not something_went_wrong:
                    tui.set_item_finished_text(
                        page.original_title,
                        rich.text.Text.from_markup(
                            "[green]:heavy_check_mark-emoji: Updated "
                            "(updated relative links)"
                        ),
                    )
                else:
                    tui.set_item_progress_label(
                        page.original_title,
                        "[red]:x: Error while updating relative links",
                    )
            else:
                tui.set_item_finished_text(
                    page.original_title,
                    rich.text.Text.from_markup(
                        "[yellow]Not updating relative links (dry run)"
                    ),
                )

            tui.set_item_progress_label(page.original_title, "")
            tui.tick_item_progress(page.original_title)

        if something_went_wrong:
            raise error


def collect_pages_to_upload(args):
    pages_to_upload: List[Page] = list()
    if not args.file_list:  # Uploading from standard input
        pages_to_upload.append(
            schneiderl_md2cf.document.get_page_data_from_lines(
                sys.stdin.readlines(),
                strip_header=args.strip_top_header,
                remove_text_newlines=args.remove_text_newlines,
                enable_relative_links=False,
            )
        )

        if not (pages_to_upload[0].title or args.title):
            error_console.log(
                "You must specify a title or have a title in the document "
                "if uploading from standard input\n"
            )
            sys.exit(1)

        if args.title:
            pages_to_upload[0].title = args.title
    else:
        for file_name in args.file_list:
            if file_name.is_dir():
                pages_to_upload += schneiderl_md2cf.document.get_pages_from_directory(
                    file_name,
                    collapse_single_pages=args.collapse_single_pages,
                    skip_empty=args.skip_empty,
                    collapse_empty=args.collapse_empty,
                    beautify_folders=args.beautify_folders,
                    remove_text_newlines=args.remove_text_newlines,
                    strip_header=args.strip_top_header,
                    use_pages_file=args.use_pages_file,
                    use_gitignore=args.use_gitignore,
                    enable_relative_links=args.enable_relative_links,
                )
            else:
                try:
                    enable_relative_links = (
                        len(args.file_list) > 1 and args.enable_relative_links
                    )
                    pages_to_upload.append(
                        schneiderl_md2cf.document.get_page_data_from_file_path(
                            file_name,
                            strip_header=args.strip_top_header,
                            remove_text_newlines=args.remove_text_newlines,
                            enable_relative_links=enable_relative_links,
                        )
                    )
                except FileNotFoundError:
                    error_console.log(f"File {file_name} does not exist\n")

        if len(pages_to_upload) == 1:
            only_page = pages_to_upload[0]

            if args.title:
                only_page.title = args.title

            # This is implicitly only truthy if relative link processing is active
            if only_page.relative_links:
                # This covers the last edge case where directory processing leaves us
                # with only one page, which we can't anticipate at startup time.
                # In this case, we have to restore all the links to their original
                # values.
                error_console.log(
                    "Relative links are ignored when there's a single page\n"
                )
                for link_data in only_page.relative_links:
                    only_page.body.replace(
                        link_data.replacement,
                        link_data.escaped_original
                        + (("#" + link_data.fragment) if link_data.fragment else ""),
                    )
                only_page.relative_links = []

    return pages_to_upload


if __name__ == "__main__":
    main()



from rich.console import Console

console = Console(log_path=False, log_time=False, emoji_variant="emoji")
error_console = Console(stderr=True, log_time=False, log_path=False)
minimal_output_console = Console(
    log_path=False,
    log_time=False,
    highlight=False,
    markup=False,
    emoji=False,
    quiet=True,
)
json_output_console = Console(
    log_path=False,
    log_time=False,
    highlight=False,
    markup=False,
    emoji=False,
    quiet=True,
)



"""
Allow checking files for ignored status in gitignore files in the repo.
"""
from pathlib import Path
from typing import List

import gitignorefile

from schneiderl_md2cf.console_output import error_console


class GitRepository:
    """
    Represents a Git repository by finding the .git folder at the root
    of the tree. Note that there are cases where .git folders may exist
    in other parts of the tree. For example in terraform repositories
    the .terraform folder may contain copies of the tree including the .git
    folder. This can be handled by initializing the GitRepository from a
    known git root. After that all files will be correctly handled by the
    is_ignored() method.
    """

    def __init__(self, repo_path: Path, use_gitignore=True):
        self.use_gitignore = use_gitignore
        self.root_dir = self._find_root_dir(repo_path) if use_gitignore else None

    @staticmethod
    def _find_root_dir(start_path: Path):
        """
        Traverse the parents of the start_path until we find a .git directory
        :param start_path: A file or directory path to start searching from
        :return: The root directory of the git repo.
        """
        fs_root = Path("/")
        p = start_path.absolute()
        if p.is_file():
            p = p.parent
        while p != fs_root:
            git_dir = p.joinpath(".git")
            if git_dir.exists() and git_dir.is_dir():
                return p
            p = p.parent
        error_console.log(
            f":warning-emoji: Directory {start_path} is not part of a git "
            f"repository: gitignore checking disabled."
        )
        return None

    def collect_gitignores(self, filepath: Path) -> List[Path]:
        """
        Collect all .gitignore files from start location to the root of the
        repository. Filepath is assumed to be a subdirectory of the git root.
        If not, an error is printed and an empty list is returned.

        :param filepath: The path to start searching for .gitignore files
        :return: List of paths to .gitignore files relevant for start_path
        """
        fs_root = Path("/")
        ret = list()

        p = filepath.absolute()
        if p.is_file():
            p = p.parent
        while p != fs_root:
            gitignore_file = p.joinpath(".gitignore")
            if gitignore_file.exists() and gitignore_file.is_file():
                ret.append(gitignore_file)
            if p == self.root_dir:
                return ret
            p = p.parent

        # if not .git directory found, we're not in a git repo and gitignore files
        # cannot be trusted.
        return list()

    def is_ignored(self, filepath: Path) -> bool:
        """
        Check if filepath is ignored in the git repository by fetching all gitignores
        in the tree down to the git root and checking all of them.

        :param filepath: Path to the file to check if it is ignored.
        :return: True if the file is ignored in any .gitignore file
        """
        if not self.use_gitignore:
            return False
        if self.root_dir is None:
            return False
        gitignores = self.collect_gitignores(filepath)
        matchers = [gitignorefile.parse(str(g)) for g in gitignores]
        return any([m(str(filepath)) for m in matchers])



import hashlib
import re
from enum import Enum
from pathlib import Path
from typing import NamedTuple

import schneiderl_md2cf.document
from schneiderl_md2cf import api

CONTENT_HASH_REGEX = re.compile(r"\[v([a-f0-9]{40})]$")


class UpsertAction(Enum):
    CREATED = 1
    UPDATED = 2
    SKIPPED = 3


class UpsertResult(NamedTuple):
    action: UpsertAction
    response: api.Bunch


# Adapted from https://stackoverflow.com/a/3431838
def get_file_sha1(file_path: Path):
    hash_sha1 = hashlib.sha1()
    with open(file_path, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash_sha1.update(chunk)
    return hash_sha1.hexdigest()


def get_parent_id_from_title(confluence, page):
    parent_page = confluence.get_page(
        title=page.parent_title,
        space_key=page.space,
        additional_expansions=[
            "space",
            "history",
            "version",
            "metadata.labels",
        ],
    )
    if parent_page is None:
        raise KeyError("The parent page could not be found")
    return parent_page.id


def upsert_page(
    confluence: api.MinimalConfluence,
    message: str,
    page: schneiderl_md2cf.document.Page,
    only_changed: bool = False,
    replace_all_labels: bool = False,
    minor_edit: bool = False,
):
    existing_page = confluence.get_page(
        title=page.title,
        space_key=page.space,
        page_id=page.page_id,
        content_type=page.content_type,
        additional_expansions=[
            "space",
            "ancestors",
            "history",
            "version",
            "metadata.labels",
        ],
    )

    # It's not mandatory to have a parent ID -- if there isn't one, the page will be a
    # top-level page in the Confluence space
    if page.parent_id is None:
        if page.parent_title is not None:
            page.parent_id = get_parent_id_from_title(confluence, page)

    page_message = message
    if only_changed:
        page_hash = page.get_content_hash()
        page_message = (
            f"{page_message} [v{page_hash}]" if page_message else f"[v{page_hash}]"
        )

    action = None
    if existing_page is None:
        existing_page = confluence.create_page(
            space=page.space,
            title=page.title,
            body=page.body,
            content_type=page.content_type,
            parent_id=page.parent_id,
            update_message=page_message,
            labels=page.labels,
        )
        action = UpsertAction.CREATED
    else:
        if not only_changed or page_needs_updating(
            page, existing_page, replace_all_labels
        ):
            existing_page = confluence.update_page(
                page=existing_page,
                body=page.body,
                parent_id=page.parent_id,
                update_message=page_message,
                labels=page.labels if replace_all_labels else None,
                minor_edit=minor_edit,
            )
            action = UpsertAction.UPDATED
        else:
            action = UpsertAction.SKIPPED

        if (
            not replace_all_labels
            and page.labels
            and labels_need_updating(page, existing_page)
        ):
            # print(f"Adding labels to page: {page.title} {page.labels}")
            confluence.add_labels(page=existing_page, labels=page.labels)

    return UpsertResult(action=action, response=existing_page)


def labels_need_updating(page, existing_page):
    if page.labels is None:
        return False

    if sorted(
        [label.name for label in existing_page.metadata.labels.results]
    ) != sorted(page.labels):
        return True


def page_needs_updating(page, existing_page, replace_all_labels):
    if page.parent_id is None and existing_page.ancestors:
        # page wants to become a top level page and was not one before
        # (top level pages only have one ancestor: the space's home page)
        return True

    if page.parent_id is not None and (
        existing_page.ancestors == []
        or page.parent_id != existing_page.ancestors[-1].id
    ):
        # page wants to change parent
        return True

    if replace_all_labels and labels_need_updating(page, existing_page):
        # print(f"Page labels have changed: {page.title} {page.labels}")
        return True
    else:
        existing_page_hash_match = CONTENT_HASH_REGEX.search(
            existing_page.version.message
        )
        if existing_page_hash_match is not None:
            original_page_hash = existing_page_hash_match.group(1)
            if original_page_hash == page.get_content_hash():
                return False

    return True


def upsert_attachment(
    confluence, attachment, existing_page, message, only_changed, page
):
    if page.file_path is not None:
        attachment_path = page.file_path.parent.joinpath(attachment)
    else:
        attachment_path = attachment

    attachment_message = message
    if only_changed:
        new_attachment_hash = get_file_sha1(attachment_path)
        attachment_message = (
            f"{attachment_message} [v{new_attachment_hash}]"
            if attachment_message
            else f"[v{new_attachment_hash}]"
        )

    existing_attachment = confluence.get_attachment(existing_page, attachment_path.name)

    action = None
    if existing_attachment is None:
        # print(f"Uploading file: {attachment_path}")
        action = UpsertAction.CREATED
        with attachment_path.open("rb") as fp:
            existing_attachment = confluence.create_attachment(
                confluence_page=existing_page, fp=fp, message=attachment_message
            )
    else:
        should_update = True
        if only_changed:
            existing_attachment_hash_match = CONTENT_HASH_REGEX.search(
                existing_attachment.version.message
            )
            if existing_attachment_hash_match is not None:
                original_attachment_hash = existing_attachment_hash_match.group(1)
                if original_attachment_hash == new_attachment_hash:
                    should_update = False
                    action = UpsertAction.SKIPPED

        if should_update:
            with attachment_path.open("rb") as fp:
                existing_attachment = confluence.update_attachment(
                    confluence_page=existing_page,
                    fp=fp,
                    existing_attachment=existing_attachment,
                    message=attachment_message,
                )
            action = UpsertAction.UPDATED

    return UpsertResult(action=action, response=existing_attachment)



from setuptools import setup

with open("README.md") as f:
    long_description = f.read()

setup(
    name="schneiderl_md2cf",
    version="0.99.99",
    packages=["schneiderl_md2cf"],
    url="https://github.com/schneiderl/md2cf",
    license="MIT",
    author="schneiderl",
    author_email="schneiderl@schneiderl.com",
    description="Convert Markdown documents to Confluence",
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    keywords="markdown confluence",
    install_requires=[
        "rich-argparse==1.0.0",
        "rich==13.0.1",
        "mistune==3.0.1",
        "chardet==5.1.0",
        "requests==2.31.0",
        "PyYAML==6.0.1",
        "gitignorefile==1.1.2",
    ],
    python_requires=">=3.7",
    entry_points={"console_scripts": ["md2cf=md2cf.__main__:main"]},
)
